import os 
import os.path as osp 
import glob 
import json 
from tqdm import tqdm
from visionsuite.utils.helpers import get_filename


def labelme2preds_json(input_dir, output_dir, class2idx):

    if not osp.exists(output_dir):
        os.makedirs(output_dir)

    json_files = glob.glob(osp.join(input_dir, '*.json'))
    
    results = {}
    for json_file in tqdm(json_files):
        filename = get_filename(json_file, False)
        img_file = osp.splitext(json_file)[0] + '.bmp'
        with open(json_file, 'r') as jf:
            anns = json.load(jf)['shapes']
            
        idx2xyxys = {}
        for ann in anns:
            if ann['label'] == 'background':
                continue
            if class2idx[ann['label']] not in idx2xyxys:
                idx2xyxys[class2idx[ann['label']]] = {'polygon': []}
                
            idx2xyxys[class2idx[ann['label']]]['polygon'].append(ann['points'])
            
        results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
                
    with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
        json.dump(results, json_file, ensure_ascii=False, indent=4)
    
if __name__ == '__main__':
    input_dir = '/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/pidnet_epochs100/test/exp/labels'
    output_dir = '/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/pidnet_epochs100/test/exp'
    classes = ['CHAMFER_MARK', 'LINE', 'MARK']
    idx2class = {idx: cls for idx, cls in enumerate(classes)}
    class2idx = {cls: idx for idx, cls in enumerate(classes)}

    labelme2preds_json(input_dir, output_dir, class2idx)





import glob 
import os
import os.path as osp
import cv2
import numpy as np
from tqdm import tqdm

input_dir = '/Data/01.Image/ctr/2_CMFB/4_TEST/gt_bgra/gt_bgra'
output_dir = '/Data/01.Image/ctr/2_CMFB/4_TEST/gt_bgra/gt_bgra_mask'

if not osp.exists(output_dir):
    os.mkdir(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.png'))

for img_file in tqdm(img_files):
    img = cv2.imread(img_file)
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    h, w, c = img.shape
    rgb_to_class = {
        (0, 0, 0): 0, 
        (255, 0, 0): 1,   # Red -> Class 0
        (0, 255, 0): 2,   # Green -> Class 1
        (0, 0, 255): 3    # Blue -> Class 2
    }
    
    class_image = np.zeros((h, w), dtype=np.uint8)
    
    for rgb, class_index in rgb_to_class.items():
        # RGB 값이 일치하는 부분에 대해 클래스 인덱스 할당
        mask = np.all(img == np.array(rgb), axis=-1)
        class_image[mask] = class_index

    
    cv2.imwrite(osp.join(output_dir, filename + '.png'), class_image)





import glob
import json
import os
import os.path as osp
import numpy as np
import cv2
from tqdm import tqdm 

img_dir = "/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/val"
json_dir = "/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/m2f_epochs100/test/exp/labels"
image_format = 'bmp'
output_dir = "/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/m2f_epochs100/test/exp/vis_labels"

if not osp.exists(output_dir):
    os.mkdir(output_dir)

# TODO: need to consider all shape-types

img_files = glob.glob(osp.join(img_dir, f'*.{image_format}'))

for img_file in tqdm(img_files):
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    json_file = osp.join(json_dir, filename + '.json')
    assert osp.exists(json_dir), ValueError(f"There is no such json-file: {json_file}")

    img = cv2.imread(img_file)
    height, width, _ = img.shape

    with open(json_file) as jf:
        anns = json.load(jf)

    for ann in anns['shapes']:
        label = ann['label']
        shape_type = ann['shape_type']
        points = ann['points']

        if shape_type == 'rectangle':
            cv2.rectangle(img, (int(points[0][0]), int(points[0][1])), (int(points[1][0]), int(points[1][1])),
                        (0, 0, 255), 5)
        elif shape_type == 'polygon':
            cv2.fillPoly(img, [np.array(points, dtype=np.int32)], color=(255, 255, 0))
        else:
            raise NotImplementedError(f"There is no such shape-type({shape_type}) considered")


    # cv2.line(img, (0, 256), (width, 256), (0, 0, 255), 2) #horizontal
    # cv2.line(img, (408, 0), (408, height), (0, 0, 255), 2) #vertical
    # cv2.line(img, (664, 0), (664, height), (0, 0, 255), 2) #vertical
    # cv2.line(img, (512, 0), (512, height), (0, 0, 255), 2)

    cv2.imwrite(osp.join(output_dir, filename + f'.{image_format}'), img)



from glob import glob 
import os.path as osp 
import os 
import json
import numpy as np
import cv2

def create_bg_crops(input_dir, output_dir, offset):

    output_dir = osp.join(output_dir, f'offset_{offset}')
    if not osp.exists(output_dir):
        os.makedirs(output_dir)

        
    img_files = glob(osp.join(input_dir, '*.bmp'))
    for img_file in img_files:
        filename = osp.split(osp.splitext(img_file)[0])[-1]
        json_file = osp.splitext(img_file)[0] + '.json'
        
        assert osp.exists(json_file), ValueError(F"There is no such json file: {json_file}")
        
        with open(json_file, 'r') as jf:
            anns = json.load(jf)['shapes']
        
        img = cv2.imread(img_file)    
        img_h, img_w, img_c = img.shape
        
        for idx, ann in enumerate(anns):
            label = ann['label']
            shape_type = ann['shape_type']
            points = np.array(ann['points'])
            
            if shape_type == 'point' or len(points) == 1:
                continue
            
            xs, ys = points[:, 0], points[:, 1]
            
            if offset == 'auto':
                _offset = max((max(xs) - min(xs))/2, (max(ys) - min(ys))/2)
            else:
                _offset = offset
            
            roi = [min(xs) - _offset, min(ys) - _offset, max(xs) + _offset, max(ys) + _offset] # x1, y1, x2, y2
            
            if roi[0] <= 0:
                roi[0] = 0
                
            if roi[1] <= 0:
                roi[1] = 0
            
            if roi[2] >= img_w:
                roi[2] = img_w
                
            if roi[3] >= img_h:
                roi[3] = img_h
                
            roi = list(map(int, roi))
            img[roi[1]:roi[3], roi[0]:roi[2]] = 0
            
        cv2.imwrite(osp.join(output_dir, filename + f'_{idx}.png'), img)
            
    
if __name__ == '__main__':
    input_dir = '/HDD/research/clustering/datasets/tenneco_outer/images'
    output_dir = '/HDD/research/clustering/datasets/tenneco_outer/bg_crops'
    offset = 'auto'
    # offset = 100

    create_bg_crops(input_dir, output_dir, offset)


from visionsuite.engines.segmentation.src.loops.build import build_loop
from visionsuite.engines.segmentation.src.datasets.build import build_dataset
from visionsuite.engines.segmentation.src.models.build import build_model
from visionsuite.engines.segmentation.utils.registry import RUNNERS
from visionsuite.engines.utils.bases import BaseTrainRunner
from visionsuite.engines.utils.callbacks import Callbacks
from .callbacks import train_callbacks


@RUNNERS.register()
class TrainRunner(BaseTrainRunner, Callbacks):
    def __init__(self, task="", name="TrainRunner"):
        BaseTrainRunner.__init__(self, task=task, name=name)
        Callbacks.__init__(self)
        
        self.add_callbacks(train_callbacks)
        
    def set_configs(self, *args, **kwargs):
        super().set_configs(mode='train', *args, **kwargs)
        
        self.run_callbacks('on_runner_set_configs')
        
    def set_variables(self):
        super().set_variables()
        
        self.run_callbacks('on_runner_set_variables')
               
    def run(self):
        super().run()
                
        self.run_callbacks('on_runner_run_start')

        # TODO: Define transform
        dataset = build_dataset(**self.args['dataset'], transform=None)
        dataset.build(**self.args['dataset'], distributed=self.args['distributed'], _logger=self.args['dataset'].get('logger', None))
        self.log_info(f"Dataset is LOADED and BUILT", self.run.__name__, __class__.__name__)
        
        model = build_model(**self.args['model'])
        model.build(**self.args['model'], 
                    num_classes=dataset.num_classes, 
                    train=self.args['train'], 
                    distributed=self.args['distributed']['use'],
                    _logger=self.args['model'].get('logger', None)
        )
        self.log_info(f"Model is LOADED and BUILT", self.run.__name__, __class__.__name__)
        
        loop = build_loop(**self.args['loop'])
        loop.build(_model=model, 
                   _dataset=dataset, 
                   _archive=self._archive, 
                   **self.args,
                   _logger=self.args['loop'].get('logger', None)
                )
        self.log_info(f"Loop is LOADED and BUILT", self.run.__name__, __class__.__name__)

        loop.run()
        
        self.run_callbacks('on_runner_run_end')

        
        # for epoch in range(self.args['start_epoch'], self.args['epochs']):
        #     if self.args['distributed']['use']:
        #         train_sampler.set_epoch(epoch)
        #     train_metric_logger = train_one_epoch(model, criterion, optimizer, data_loader, lr_scheduler, self.args['device'], epoch, self.args['print_freq'], scaler)
        #     confmat, val_metric_logger = evaluate(model, data_loader_test, device=self.args['device'], num_classes=num_classes)
            
        #     monitor.log({"learning rate": train_metric_logger.meters['lr'].value})
        #     monitor.log({"train avg loss": train_metric_logger.meters['loss'].avg})
        #     for key, val in confmat.values.items():
        #         if 'acc' in key:
        #             monitor.log({key: val})
        #         if 'iou' in key:
        #             monitor.log({key: val})
        #     monitor.save()



import os 
import os.path as osp 
import glob 
import json 
import cv2
from tqdm import tqdm
from shutil import copyfile
import numpy as np
from visionsuite.utils.helpers import get_filename

from visionsuite.utils.dataset.formats.labelme.utils import get_mask_from_labelme


def labelme2mask(input_dir, output_dir, class2label, input_format, modes, width=None, height=None, vis=False, output_format='bmp', rois=[[]]):

    if not osp.exists(output_dir):
        os.makedirs(output_dir)

    for mode in modes:
        _output_dir = osp.join(output_dir, mode)
        if not osp.exists(_output_dir):
            os.mkdir(_output_dir)
            
        if vis:
            vis_dir = osp.join(_output_dir, 'vis')
            if not osp.exists(vis_dir):
                os.mkdir(vis_dir)
            
        json_files = glob.glob(osp.join(input_dir, mode, '*.json'))
        
        for json_file in tqdm(json_files, desc=mode):
            img_file = osp.splitext(json_file)[0] + f'.{input_format}'
            img = cv2.imread(img_file)
            filename = get_filename(json_file, False)
            with open(json_file, 'r') as jf:
                anns = json.load(jf)
                
            if width is None and height is None:
                width, height = anns['imageWidth'], anns['imageHeight']
                
            if rois == [[]]:
                rois = [[0, 0, width, height]]
            mask = get_mask_from_labelme(json_file, width, height, class2label, 
                                        format='opencv')
            import numpy as np
            mask_output_dir = osp.join(_output_dir, 'masks')
            if not osp.exists(mask_output_dir):
                os.mkdir(mask_output_dir)
                        
            image_output_dir = osp.join(_output_dir, 'images')
            if not osp.exists(image_output_dir):
                os.mkdir(image_output_dir)
                
            for roi in rois:
                roi_img = img[roi[1]:roi[3], roi[0]:roi[2]]
                roi_mask = mask[roi[1]:roi[3], roi[0]:roi[2]]
                
                cv2.imwrite(osp.join(mask_output_dir, filename + f'.{output_format}'), roi_mask)
                cv2.imwrite(osp.join(image_output_dir, filename + f'.{output_format}'), roi_img)
                # copyfile(img_file, osp.join(image_output_dir, filename + f'.{output_format}'))
                
                if vis:
                    import numpy as np
                    import imgviz 
                    
                    vis_img = np.zeros((roi[3] - roi[1], (roi[2] - roi[0])*2, 3))
                    vis_img[:, :roi[2] - roi[0], :] = roi_img
                    color_map = imgviz.label_colormap(50)
                    vis_img[:, roi[2] - roi[0]:, :] = color_map[roi_mask.astype(np.uint8)].astype(np.uint8) 
                    
                    cv2.imwrite(osp.join(vis_dir, filename + '.png'), vis_img)
        
            
if __name__ == '__main__':
    input_dir = '/storage/projects/Tenneco/Metalbearing/OUTER/250211/split_dataset'
    output_dir = '/storage/projects/Tenneco/Metalbearing/OUTER/250211/split_mask_dataset'
    # input_dir = '/HDD/datasets/projects/Tenneco/Metalbearing_outer/split_dataset_'
    # output_dir = '/HDD/datasets/projects/Tenneco/Metalbearing_outer/split_mask_dataset_'
    class2label = {'chamfer_mark': 1, 'line': 2, 'mark': 3}
    # class2label = {"stabbed": 1}
    # class2label = {'CHAMFER_MASK': 1, 'LINE': 2, 'MASK': 3}
    modes = ['train', 'val']
    # width, height = 1120, 768
    width, height = None, None
    vis = True
    output_format = 'bmp'
    input_format = 'bmp'
    roi = [[220, 60, 1340, 828]]

    labelme2mask(input_dir, output_dir, class2label, input_format, modes, width, height, vis, output_format, roi)


    # input_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/tmp_patch_dataset'
    # output_dir = '/HDD/datasets/projects/LX/24.11.28_2/datasets_wo_vertical/datasets/tmp_patch_dataset/vis'
    # # class2label = {'tear': 1}
    # # class2label = {'scratch': 1}
    # # class2label = {'scratch': 1, 'tear': 2}
    # # class2label = {'scratch': 1, 'tear': 2, 'stabbed': 3}
    # class2label = {'timber': 1, 'screw': 2}
    # modes = ['./']
    # width, height = 1024, 1024
    # vis = True
    # output_format = 'png'
    # input_format = 'jpg'

    # labelme2mask(input_dir, output_dir, class2label, input_format, modes, width, height, vis, output_format)





from visionsuite.utils.dataset.formats.labelme.utils import *

import glob 
import os
import os.path as osp
import cv2
import numpy as np
from tqdm import tqdm

input_dir = '/Data/01.Image/ctr/2_CMFB/4_TEST/gt_bgra/gt_bgra_mask'

img_files = glob.glob(osp.join(input_dir, '*.png'))

for img_file in tqdm(img_files):
    img = cv2.imread(img_file, -1)   
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    h, w = img.shape
    
    _labelme = init_labelme_json(filename + ".png", w, h)
    _labelme = get_points_from_image(
                                img,
                                ['background', 'ballstud_s', 'taper_s', 'ballstud_head_s'],
                                [0, 0],
                                [0, 0],
                                _labelme,
                                50,
                            )
    
    with open(os.path.join(input_dir, filename + ".json"), "w") as jsf:
        json.dump(_labelme, jsf)


import glob 
import os
import os.path as osp

import imgviz
import json
import cv2
import numpy as np
from tqdm import tqdm
import pandas as pd
from visionsuite.utils.visualizers.vis_seg import vis_seg
from athena.src.tasks.segmentation.frameworks.tensorflow.models.tf_model_v2 import TFModelv2


def get_mask_from_pred(pred, conf=0.5, contour_thres=10):
    mask = (pred > conf).astype(np.uint8)
    contours, _ = cv2.findContours(mask, cv2.RETR_EXTERNAL, cv2.CHAIN_APPROX_NONE)
    points = []
    for contour in contours:
        if len(contour) < contour_thres:
            pass
        else:
            epsilon = 0.001 * cv2.arcLength(contour, True)
            approx = cv2.approxPolyDP(contour, epsilon, True)[:, 0, :].tolist()
            points.append(approx)

    return mask, points

compare_mask = False
weights = '/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/tf_deeplabv3plus_epochs200/train/weights/last_weights.h5'
model = TFModelv2(model_name='deeplabv3plus', backbone='efficientnetb3', backbone_weights='imagenet', backbone_trainable=True, 
                  batch_size=1, width=1120, height=768, channel=3, num_classes=4,
                  weights=weights)
blob_confidence = 0.8
classes = ["CHAMFER_MARK", "LINE", "MARK"]
idx2class = {idx: cls for idx, cls in enumerate(classes)}
_classes = ["CHAMFER_MARK", "LINE", "MARK"]
_idx2class = {idx: cls for idx, cls in enumerate(_classes)}
input_dir = '/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/val'
json_dir = '/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/val'
output_dir = f'/DeepLearning/etc/_athena_tests/benchmark/tenneco/outer/outputs/tf_deeplabv3plus_epochs200/test/exp_conf{blob_confidence}'
roi = [220, 60, 1340, 828]

if not osp.exists(output_dir):
    os.makedirs(output_dir)

img_files = glob.glob(osp.join(input_dir, '*.bmp'))

results = {}
compare = {}
for img_file in tqdm(img_files):
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file)
    
    if roi is not None:
        img = img[roi[1]:roi[3], roi[0]:roi[2]]
    
    preds = model(img[np.newaxis, :])[0].numpy()
    
    idx2xyxys = {}
    for idx in idx2class.keys():
        mask, points = get_mask_from_pred(preds[:, :, idx + 1], conf=blob_confidence)
    
        for _points in points:
            for _point in _points:
                _point[0] += roi[0]
                _point[1] += roi[1]
        
            if idx not in idx2xyxys:
                idx2xyxys[idx] = {'polygon': []}
            
            idx2xyxys[idx]['polygon'].append(_points)
        
    # if _classes is not None:
    #     new_idx2xyxys = {}
    #     for idx, _cls in enumerate(_classes):
    #         for jdx, cls in enumerate(classes):
    #             if cls == _cls:
    #                 new_idx2xyxys[idx] = idx2xyxys[jdx]
        
    #     idx2xyxys = new_idx2xyxys
    #     idx2class = _idx2class    
    
    color_map = imgviz.label_colormap()[1:len(idx2class) + 1]
    results.update({filename: {'idx2xyxys': idx2xyxys, 'idx2class': idx2class, 'img_file': img_file}})
    
    if compare_mask:
        _compare = vis_seg(img_file, idx2xyxys, idx2class, output_dir, color_map, seg_type='semantic', 
                           json_dir=json_dir, compare_mask=compare_mask)
        _compare.update({"img_file": img_file})
        compare.update({filename: _compare})
    else:
        vis_seg(img_file, idx2xyxys, idx2class, output_dir, color_map, json_dir=json_dir, compare_mask=compare_mask)
    
with open(osp.join(output_dir, 'preds.json'), 'w', encoding='utf-8') as json_file:
    json.dump(results, json_file, ensure_ascii=False, indent=4)

if compare_mask:
    with open(osp.join(output_dir, 'diff.json'), 'w', encoding='utf-8') as json_file:
        json.dump(compare, json_file, ensure_ascii=False, indent=4)
    
    df_compare = pd.DataFrame(compare)
        
    df_compare_pixel = df_compare.loc['diff_pixel'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_pixel.csv'))
    
    df_compare_pixel = df_compare.loc['diff_iou'].T
    df_compare_pixel.fillna(0, inplace=True)
    df_compare_pixel.to_csv(osp.join(output_dir, 'diff_iou.csv'))


import torch
import numpy as np
import shutil
from PIL import Image
from pathlib import Path
from collections import defaultdict
from transformers import CLIPProcessor, CLIPModel, Dinov2Model
import argparse
from tqdm.auto import tqdm
from annoy import AnnoyIndex
from scipy.cluster.hierarchy import linkage, fcluster
from scipy.spatial.distance import squareform
import os
import os.path as osp
import cv2 
from torchvision import transforms


def generate_dinov2_embeddings(image_ids, image_paths, model, transform, device, batch_size, regenerate, embeddings_file):
    if not regenerate and embeddings_file.exists():
        return [], np.load(embeddings_file)

    embeddings = []
    damaged_image_ids = []

    with torch.no_grad():
        for i in range(0, len(image_ids), batch_size):
            batch_image_ids = image_ids[i:i+batch_size]
            batch_images = []

            for image_id in batch_image_ids:
                try:
                    image = Image.open(image_paths[image_id]).convert("RGB")
                    image = transform(image).unsqueeze(0).to(device)
                    batch_images.append(image)
                except Exception:
                    damaged_image_ids.append(image_id)

            if batch_images:
                batch_images = torch.cat(batch_images, dim=0)
                features = model(batch_images).last_hidden_state[:, 0, :].cpu().numpy()  # CLS token 사용
                embeddings.append(features)

    all_embeddings = np.concatenate(embeddings, axis=0)
    np.save(embeddings_file, all_embeddings)
    return damaged_image_ids, all_embeddings


def get_embeddings(model_dict, device, image_directory, batch_size, regenerate_embeddings, embeddings_file):
    
    allowed_extensions = {".jpeg", ".jpg", ".png", ".webp", '.bmp'}
    images_to_paths, all_image_ids = get_images_to_paths(image_directory, allowed_extensions)
    
    if model_dict['model_name'] == 'clip':
        model = CLIPModel.from_pretrained(model_dict['ckpt']).to(device)
        processor = CLIPProcessor.from_pretrained(model_dict['ckpt'])

        images_to_paths, all_image_ids = get_images_to_paths(image_directory, allowed_extensions)
        damaged_image_ids, all_embeddings = generate_embeddings(all_image_ids, images_to_paths, model, 
                                                                processor, device, batch_size, regenerate_embeddings, embeddings_file)

    elif model_dict['model_name'] == 'dinov2':
        model = Dinov2Model.from_pretrained(model_dict['ckpt']).to(device)
        model.eval()

        # DINOv2 이미지 변환 정의
        transform = transforms.Compose([
            transforms.Resize((224, 224)),
            transforms.ToTensor(),
            transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
        ])

        allowed_extensions = {".jpeg", ".jpg", ".png", ".webp", ".bmp"}

        damaged_image_ids, all_embeddings = generate_dinov2_embeddings(
            all_image_ids, images_to_paths, model, transform, device, batch_size, regenerate_embeddings, embeddings_file
        )

    return damaged_image_ids, all_embeddings, all_image_ids, images_to_paths

def process_images(image_directory, model_dict, threshold, batch_size, output_dir, device, offset):
    
    output_dir = osp.join(output_dir, f"{model_dict['model_name']}_{model_dict['ckpt']}", f"offset_{offset}", f"th_{threshold}")
    if not osp.exists(output_dir):
        os.makedirs(output_dir)
    
    image_directory = Path(image_directory)
    device = "cuda" if torch.cuda.is_available() else "cpu"

    embeddings_file = image_directory / 'embeddings.npy'
    regenerate_embeddings = check_and_load_embeddings(embeddings_file)

    damaged_image_ids, all_embeddings, all_image_ids, images_to_paths = get_embeddings(model_dict, device, image_directory, batch_size,
                                                       regenerate_embeddings, embeddings_file)

    # if regenerate_embeddings:
    #     np.save(embeddings_file, all_embeddings)

    print("Building Annoy index...")
    annoy_index = build_annoy_index(all_embeddings)

    print("Computing distance matrix...")
    distances = compute_distance_matrix(all_embeddings, annoy_index)

    print("Applying hierarchical clustering...")
    labels = apply_clustering(distances, threshold)
    print(len(np.unique(labels)), len(labels), labels)

    image_id_clusters = build_image_clusters(all_image_ids, labels)
    print(image_id_clusters)

    for idx, image_id_cluster in image_id_clusters.items():
        mosaic = np.zeros((1024, 1024, 3))
        cnt, jdx = 0, 0
        for image in image_id_cluster:
            img = cv2.imread(images_to_paths[image])
            
            img = cv2.resize(img, (256, 256))
            mosaic[int(256*(cnt//4)):int(256*(cnt//4 + 1)), int(256*(cnt%4)):int(256*(cnt%4 + 1))] += img
            cnt += 1
            if cnt > 15:
                cv2.imwrite(osp.join(output_dir, f"{idx}_{jdx}.png"), mosaic)
                jdx += 1
                cnt = 0
                mosaic = np.zeros((1024, 1024, 3))
                
        cv2.imwrite(osp.join(output_dir, f"{idx}_{jdx}.png"), mosaic)
        
def check_and_load_embeddings(embeddings_file):
    # if embeddings_file.exists():
    #     use_existing_embeddings = input("Embeddings file found. Do you want to use existing embeddings? (Y/N) ").strip().lower()
    #     if use_existing_embeddings in ('', 'y', 'yes'):
    #         print("Loading embeddings from file...")
    #         all_embeddings = np.load(embeddings_file)
    #         return False
    # return True
    return True

# Get the paths of all images in the given directory and return the image ids and their paths
def get_images_to_paths(image_directory, allowed_extensions):
    images_to_paths = {
        image_path.stem: image_path
        for image_path in image_directory.iterdir()
        if image_path.suffix.lower() in allowed_extensions
    }
    return images_to_paths, list(images_to_paths.keys())

# Generate CLIP embeddings for all images, handling damaged images if any
def generate_embeddings(all_image_ids, images_to_paths, model, processor, device, batch_size, regenerate_embeddings, embeddings_file):
    if not regenerate_embeddings:
        return set(), np.load(embeddings_file)

    damaged_image_ids, all_embeddings = set(), []
    progress_bar = tqdm(total=len(all_image_ids), desc="Generating CLIP embeddings")

    for i in range(0, len(all_image_ids), batch_size):
        batch_image_ids, batch_images = process_image_batch(all_image_ids, i, batch_size, images_to_paths, damaged_image_ids)
        inputs = processor(images=batch_images, return_tensors="pt", padding=True).to(device)

        with torch.no_grad():
            outputs = model.get_image_features(**inputs)

        all_embeddings.extend(outputs.cpu().numpy())
        progress_bar.update(len(batch_image_ids))

    progress_bar.close()
    return damaged_image_ids, all_embeddings

# Process a batch of images, returning their ids and loaded images, while identifying damaged images
def process_image_batch(all_image_ids, start_idx, batch_size, images_to_paths, damaged_image_ids):
    batch_image_ids = all_image_ids[start_idx: start_idx + batch_size]
    batch_images = []

    for image_id in batch_image_ids:
        try:
            image = Image.open(images_to_paths[image_id])
            image.load()
            batch_images.append(image)
        except OSError:
            print(f"\nError processing image {images_to_paths[image_id]}, marking as corrupted.")
            damaged_image_ids.add(image_id)

    return batch_image_ids, batch_images

# Build an Annoy index using the generated CLIP embeddings
def build_annoy_index(all_embeddings):
    embeddings = np.array(all_embeddings)
    n_dimensions = embeddings.shape[1]

    annoy_index = AnnoyIndex(n_dimensions, "angular")
    for i, embedding in enumerate(embeddings):
        annoy_index.add_item(i, embedding)

    annoy_index.build(1000)
    return annoy_index

# Compute the distance matrix of the embeddings using the Annoy index
def compute_distance_matrix(all_embeddings, annoy_index):
    n = len(all_embeddings)
    distances = []

    for i in range(n):
        for j in range(i + 1, n):
            distance = annoy_index.get_distance(i, j)
            distances.append(distance)

    return distances

# Apply hierarchical clustering on the computed distance matrix with the given threshold
def apply_clustering(distances, threshold):
    condensed_distances = np.array(distances)
    Z = linkage(condensed_distances, method='average', optimal_ordering=True)
    return fcluster(Z, t=threshold, criterion='distance')

# Build clusters of image ids based on the clustering labels
def build_image_clusters(all_image_ids, labels):
    image_id_clusters = defaultdict(set)

    for image_id, cluster_label in zip(all_image_ids, labels):
        image_id_clusters[cluster_label].add(image_id)

    return image_id_clusters

def main():

    input_dir = '/HDD/research/clustering/datasets/tenneco_outer/bg_crops'
    output_dir = '/HDD/research/clustering/datasets/tenneco_outer/outputs_bg'
    threshold = 0.7
    batch_size = 1
    device = 'cuda:0'
    # model_dict = {'model_name': 'clip', 'ckpt': 'openai/clip-vit-large-patch14-336', }
    model_dict = {'model_name': 'dinov2', 'ckpt': 'facebook/dinov2-large', }
    offset = 'auto'
    # offset = 100

    process_images(osp.join(input_dir, f'offset_{offset}'), model_dict, threshold, 
                   batch_size, output_dir, device, offset)

if __name__ == "__main__":
    main()


import os.path as osp
import cv2 
import numpy as np
import warnings
from visionsuite.utils.metrics.iou import get_iou
from shapely.geometry import (GeometryCollection, LineString, MultiLineString, MultiPoint, MultiPolygon, Point, Polygon,
                              mapping)
from shapely.ops import polygonize, unary_union
from visionsuite.utils.helpers import get_text_coords

def handle_self_intersection(points):
    new_points = []
    line = LineString([[int(x), int(y)] for x, y in points + [points[0]]])

    polygons = list(polygonize(unary_union(line)))

    if len(polygons) > 1:
        print("The line is forming polygons by intersecting itself")
        for polygon in polygons:
            polygon = [list(item) for item in mapping(polygon)['coordinates'][0]]
            new_points.append(polygon[:-1])
    else:
        return [points]

    return new_points


def get_key_by_value(dictionary, value):
    for key, val in dictionary.items():
        if val == value:
            return key
    return None

def vis_seg(img_file, idx2xyxys, idx2class, output_dir, color_map, 
            seg_type='semantic', 
            json_dir=None, compare_mask=True, 
            iou_threshold=0.2, font_scale=1, line_width=1, draw_rect=True):
   
    class2idx = {cls: idx for idx, cls in idx2class.items()}
    
    filename = osp.split(osp.splitext(img_file)[0])[-1]
    img = cv2.imread(img_file)
    height, width, channel = img.shape
    vis_mask = np.zeros((height, width, channel))
    if compare_mask:
        diff_dict = {}
        points_dict = {'gt': {}, 'pred': {}}
        compr_pred_mask = {cls: np.zeros((height, width)) for cls in idx2class.values()}
    else:
        compr_pred_mask = None
    compr_gt_mask, vis_gt = None, None
    
    origin = 25, 25
    font = cv2.FONT_HERSHEY_SIMPLEX
    text_ori = np.zeros((50, width, 3), np.uint8)
    text_pred = np.zeros((50, width, 3), np.uint8)
    text_legends = np.zeros((50, 300, 3), np.uint8)
    cv2.putText(text_ori, "Original", origin, font, 0.6, (255, 255, 255), 1)
    cv2.putText(text_pred, "Predicted", origin, font, 0.6, (255, 255, 255), 1)
    cv2.putText(text_legends, "Legends", origin, font, 0.6, (255, 255, 255), 1)
    
    if json_dir is not None:
        import json
        json_file = osp.join(json_dir, filename + '.json')
        vis_gt = np.zeros((height, width, channel))
        if compare_mask:
            compr_gt_mask = {cls: np.zeros((height, width)) for cls in idx2class.values()}
        text_gt = np.zeros((50, width, 3), np.uint8)
        cv2.putText(text_gt, "GT", origin, font, 0.6, (255, 255, 255), 1)
        
        with open(json_file, 'r') as jf:
            anns = json.load(jf)['shapes']
            
        for idx, cls in idx2class.items():
            for ann in anns:
                label = ann['label']
                if label != cls:
                    continue
                points = ann['points']
                shape_type = ann['shape_type']
                
                if len(points) > 2:
                    cv2.fillConvexPoly(vis_gt, np.array(points, dtype=np.int32), 
                                color=tuple(map(int, color_map[int(get_key_by_value(idx2class, label))])))
                    if font_scale:
                        cv2.putText(vis_gt, label, get_text_coords(points, width, height, offset_h=10), cv2.FONT_HERSHEY_SIMPLEX, font_scale, tuple(map(int, color_map[-1])), 3)
                    if compare_mask:
                        cv2.fillConvexPoly(compr_gt_mask[label], np.array(points, dtype=np.int32), 
                                color=1)
                        # cv2.putText(compr_gt_mask[label], label, get_text_coords(points, width, height, offset_h=10), cv2.FONT_HERSHEY_SIMPLEX, font_scale, tuple(map(int, color_map[-1])), 3)
                        if label in points_dict['gt']:
                            points_dict['gt'][label].append(points)
                        else:
                            points_dict['gt'].update({label: [points]})
                else:
                    warnings.warn(f"The points is {points} with {shape_type}")
                
        vis_gt = vis_gt.astype(np.uint8)
        vis_gt = cv2.addWeighted(img, 0.4, vis_gt, 0.6, 0)
        vis_gt = cv2.vconcat([text_gt, vis_gt])
            
    if seg_type == 'semantic':
        for idx, cls in idx2class.items():
            if class2idx[cls] in idx2xyxys:
                pred = idx2xyxys[class2idx[cls]]
                for mask in pred['polygon']:
                    points = np.array(mask, dtype=np.int32)
                    if points.size != 0:
                        if font_scale:
                            cv2.putText(vis_mask, f"{cls}", 
                                        get_text_coords(points, width, height, offset_h=10), cv2.FONT_HERSHEY_SIMPLEX, font_scale, 
                                        tuple(map(int, color_map[int(idx)])), 3)
                        cv2.fillConvexPoly(vis_mask, points, color=tuple(map(int, color_map[int(idx)])))
                        # cv2.fillPoly(vis_mask, [points], color=tuple(map(int, color_map[int(idx)])))
                        if compare_mask:
                            # cv2.putText(compr_pred_mask[idx], f"{idx}", get_text_coords(points, width, height, offset_h=30), cv2.FONT_HERSHEY_SIMPLEX, font_scale, tuple(map(int, color_map[int(cls)])), 3)
                            cv2.fillConvexPoly(compr_pred_mask[cls], points, color=1)

                            if cls in points_dict['pred']:
                                points_dict['pred'][cls].append(points.tolist())
                            else:
                                points_dict['pred'].update({cls: [points.tolist()]})
                
    elif seg_type == 'instance':
        for idx, cls in idx2class.items():
            if class2idx[cls] in idx2xyxys:
                pred = idx2xyxys[class2idx[cls]]
                for mask, box, conf in zip(pred['polygon'], pred['box'], pred['confidence']):
                    points = np.array(mask, dtype=np.int32)
                    if points.size != 0:
                        if font_scale:
                            cv2.putText(vis_mask, f"{cls} {conf:.2f}", 
                                        get_text_coords(points, width, height, offset_h=10), cv2.FONT_HERSHEY_SIMPLEX, font_scale, 
                                        tuple(map(int, color_map[int(idx)])), 3)
                        cv2.fillConvexPoly(vis_mask, points, color=tuple(map(int, color_map[int(idx)])))
                        # cv2.fillPoly(vis_mask, [points], color=tuple(map(int, color_map[int(idx)])))
                        if compare_mask:
                            # cv2.putText(compr_pred_mask[idx], f"{idx}", get_text_coords(points, width, height, offset_h=30), cv2.FONT_HERSHEY_SIMPLEX, font_scale, tuple(map(int, color_map[int(cls)])), 3)
                            cv2.fillConvexPoly(compr_pred_mask[cls], points, color=1)

                            if cls in points_dict['pred']:
                                points_dict['pred'][cls].append(points.tolist())
                            else:
                                points_dict['pred'].update({cls: [points.tolist()]})
                    
                    if draw_rect and len(box) != 0:
                        cv2.rectangle(vis_mask, (int(box[0][0]), int(box[0][1])), (int(box[1][0]), int(box[1][1])), 
                                        tuple(map(int, color_map[int(idx)])), line_width + 1)
                        if font_scale:
                            cv2.putText(vis_mask, label, get_text_coords(box, width, height), cv2.FONT_HERSHEY_SIMPLEX, font_scale, 
                                    tuple(map(int, color_map[int(idx)])), line_width)
    
    else:
        raise ValueError(f"There is no such seg-type: {seg_type}")

    vis_img = cv2.vconcat([text_ori, img])

    vis_mask = vis_mask.astype(np.uint8)
    vis_mask = cv2.addWeighted(img, 0.4, vis_mask, 0.6, 0)
    vis_mask = cv2.vconcat([text_pred, vis_mask])

    vis_legend = np.zeros((height, 300, channel), dtype="uint8")
    for idx, color in enumerate(color_map):
        color = [int(c) for c in color]
        cv2.putText(vis_legend, idx2class[idx], (5, (idx * 25) + 17), cv2.FONT_HERSHEY_SIMPLEX, 0.5, (0, 0, 255), 2)
        cv2.rectangle(vis_legend, (150, (idx * 25)), (300, (idx * 25) + 25), tuple(color), -1)
    vis_legend = cv2.vconcat([text_legends, vis_legend])

    if vis_gt is None:
        vis_res = cv2.hconcat([vis_img, vis_mask, vis_legend])
    else:
        vis_res = cv2.hconcat([vis_img, vis_gt, vis_mask, vis_legend])

    cv2.imwrite(osp.join(output_dir, filename + '.png'), vis_res)
        
    if compare_mask:
        for idx, (key, val) in enumerate(compr_gt_mask.items()):
            origin = 25, 25
            font = cv2.FONT_HERSHEY_SIMPLEX
            text_compr = np.zeros((50, val.shape[1]))
            if idx != 0:
                _vis_compr = val - compr_pred_mask[key]
                diff = np.sum(np.abs(vis_compr))
                cv2.putText(text_compr, f'{key} ({str(diff)})', origin, font, 0.6, (255, 255, 255), 1)
                _vis_compr = cv2.vconcat([text_compr, _vis_compr])
                vis_compr = cv2.hconcat([vis_compr, _vis_compr])
                diff_dict[key] = diff
            else:
                vis_compr = val - compr_pred_mask[key]
                diff = np.sum(np.abs(vis_compr))
                cv2.putText(text_compr, f'{key} ({str(diff)})', origin, font, 0.6, (255, 255, 255), 1)
                vis_compr = cv2.vconcat([text_compr, vis_compr])
                diff_dict[key] = diff
        
        vis_compr = 255*np.abs(vis_compr)
        cv2.imwrite(osp.join(output_dir, filename + '_diff.png'), vis_compr)
            
        ious = {}
        for gt_cls, gt_points in points_dict['gt'].items():
            for pred_cls, pred_points in points_dict['pred'].items():
                if gt_cls == pred_cls:
                    for gt_point in gt_points:
                        for pred_point in pred_points:
                            try:
                                iou = get_iou(gt_point, pred_point, filename=filename)
                                if iou > iou_threshold:
                                    if gt_cls in ious:
                                        ious[gt_cls].append(iou)
                                    else:
                                        ious[gt_cls] = [iou]
                                
                            except:
                                _gt_points = handle_self_intersection(gt_point)
                                
                                for _gt_point in _gt_points:
                                    iou = get_iou(_gt_point, pred_point, filename=filename)
                                    if iou > iou_threshold:
                                        if gt_cls in ious:
                                            ious[gt_cls].append(iou)
                                        else:
                                            ious[gt_cls] = [iou]
            
        return {'diff_pixel': diff_dict, 'diff_iou': ious}


import os
import cv2
import numpy as np
from sklearn.cluster import KMeans
from sklearn.decomposition import PCA
from sklearn.metrics import silhouette_score
import torch
import torchvision.models as models
import torchvision.transforms as transforms
from PIL import Image
import os.path as osp

# Step 1: Preprocess images and extract features

def preprocess_and_extract_features(image_folder, model):
    preprocess = transforms.Compose([
        transforms.Resize((224, 224)),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225]),
    ])

    features = []
    image_paths = []

    for filename in os.listdir(image_folder):
        if filename.endswith(('.jpg', '.jpeg', '.png', '.bmp')):
            image_path = os.path.join(image_folder, filename)
            image = Image.open(image_path).convert('RGB')
            input_tensor = preprocess(image).unsqueeze(0)
            
            with torch.no_grad():
                feature = model(input_tensor)
            
            features.append(feature.squeeze().numpy())
            image_paths.append(image_path)

    return np.array(features), image_paths

# Step 2: Apply PCA for dimensionality reduction

def apply_pca(features, n_components=18):
    pca = PCA(n_components=n_components)
    return pca.fit_transform(features)

# Step 3: Perform K-means clustering

def perform_kmeans(features, max_clusters=10):
    best_n_clusters = 2
    best_score = -1

    for n_clusters in range(2, min(max_clusters + 1, len(features))):
        kmeans = KMeans(n_clusters=n_clusters, random_state=42)
        cluster_labels = kmeans.fit_predict(features)
        score = silhouette_score(features, cluster_labels)
        
        if score > best_score:
            best_score = score
            best_n_clusters = n_clusters

    kmeans = KMeans(n_clusters=best_n_clusters, random_state=42)
    return kmeans.fit_predict(features)

def select_representatives(features, cluster_labels, image_paths, n_representatives=3):
    representatives = []
    for cluster in range(max(cluster_labels) + 1):
        cluster_indices = np.where(cluster_labels == cluster)[0]
        cluster_features = features[cluster_indices]
        cluster_center = np.mean(cluster_features, axis=0)
        
        distances = np.linalg.norm(cluster_features - cluster_center, axis=1)
        sorted_indices = np.argsort(distances)
        
        for i in range(min(n_representatives, len(cluster_indices))):
            rep_index = cluster_indices[sorted_indices[i]]
            representatives.append(image_paths[rep_index])

    return representatives

# def select_representatives(features, cluster_labels, image_paths, min_total_representatives=3, max_per_cluster=5):
#     representatives = []
#     for cluster in range(max(cluster_labels) + 1):
#         cluster_indices = np.where(cluster_labels == cluster)[0]
#         cluster_features = features[cluster_indices]
#         cluster_center = np.mean(cluster_features, axis=0)
        
#         distances = np.linalg.norm(cluster_features - cluster_center, axis=1)
#         sorted_indices = np.argsort(distances)
        
#         n_representatives = min(max_per_cluster, len(cluster_indices))
#         for i in range(n_representatives):
#             rep_index = cluster_indices[sorted_indices[i]]
#             representatives.append(image_paths[rep_index])

#     # If we don't have enough representatives, add more from larger clusters
#     while len(representatives) < min_total_representatives:
#         for cluster in range(max(cluster_labels) + 1):
#             if len(representatives) >= min_total_representatives:
#                 break
#             cluster_indices = np.where(cluster_labels == cluster)[0]
#             if len(cluster_indices) > len(representatives) // max(cluster_labels) + 1:
#                 for idx in cluster_indices:
#                     if image_paths[idx] not in representatives:
#                         representatives.append(image_paths[idx])
#                         break

#     return representatives


# Main execution

if __name__ == "__main__":
    input_dir = '/HDD/research/clustering/datasets/tenneco_outer/crop'
    output_dir = '/HDD/research/clustering/datasets/tenneco_outer/outputs_v2'

    offset = 'auto'

    input_dir = osp.join(input_dir, f'offset_{offset}')

    # Load pre-trained ResNet model
    model = models.resnet50(pretrained=True)
    model = torch.nn.Sequential(*list(model.children())[:-1])  # Remove the last fully connected layer
    model.eval()

    # Extract features
    features, image_paths = preprocess_and_extract_features(input_dir, model)

    # Apply PCA
    reduced_features = apply_pca(features)

    # Perform clustering
    cluster_labels = perform_kmeans(reduced_features)

    # Select representatives
    representatives = select_representatives(reduced_features, cluster_labels, image_paths)

    # Copy representative images to output folder
    os.makedirs(output_dir, exist_ok=True)
    for rep_path in representatives:
        filename = os.path.basename(rep_path)
        output_path = os.path.join(output_dir, filename)
        cv2.imwrite(output_path, cv2.imread(rep_path))

    print(f"Selected {len(representatives)} representative images out of {len(image_paths)} total images.")
    print(f"Representative images saved to {output_dir}")



import torch
import torchvision.models as models
import torchvision.transforms as transforms
import cv2
import numpy as np
from PIL import Image
import matplotlib.pyplot as plt

class GradCAM:
    def __init__(self, model, target_layer):
        self.model = model
        self.target_layer = target_layer
        self.gradients = None
        self.activations = None
        
        def save_gradient(module, grad_input, grad_output):
            self.gradients = grad_output[0]
        
        def save_activation(module, input, output):
            self.activations = output
        
        target_layer.register_forward_hook(save_activation)
        target_layer.register_backward_hook(save_gradient)
    
    def generate_cam(self, input_tensor, target_class):
        self.model.zero_grad()
        output = self.model(input_tensor)
        class_score = output[0, target_class]
        class_score.backward()
        
        # Compute Grad-CAM
        gradients = self.gradients.cpu().data.numpy()
        activations = self.activations.cpu().data.numpy()
        weights = np.mean(gradients, axis=(2, 3), keepdims=True)
        cam = np.sum(weights * activations, axis=1)
        
        cam = np.maximum(cam, 0)
        cam = cam[0]
        cam = cv2.resize(cam, (224, 224))
        cam = cam - np.min(cam)
        cam = cam / np.max(cam)
        
        return cam

# 2. 이미지 전처리 함수
def preprocess_image(img_path):
    transform = transforms.Compose([
        transforms.Resize((224, 224)),
        transforms.ToTensor(),
        transforms.Normalize(mean=[0.485, 0.456, 0.406], std=[0.229, 0.224, 0.225])
    ])
    image = Image.open(img_path).convert('RGB')
    image_tensor = transform(image).unsqueeze(0)
    return image_tensor, np.array(image)

# 3. CAM 시각화 함수
def overlay_cam_on_image(img, cam):
    heatmap = cv2.applyColorMap(np.uint8(255 * cam), cv2.COLORMAP_JET)
    heatmap = cv2.cvtColor(heatmap, cv2.COLOR_BGR2RGB)
    img = cv2.resize(img, (224, 224))
    overlayed = 0.5 * heatmap + 0.5 * img
    overlayed = np.clip(overlayed, 0, 255).astype(np.uint8)
    return overlayed

# 4. 실행 코드
def main(img_path):
    # 모델 및 Grad-CAM 설정
    model = models.resnet50(pretrained=True)
    model.eval()
    target_layer = model.layer4[-1]
    grad_cam = GradCAM(model, target_layer)
    
    # 이미지 로드 및 전처리
    image_tensor, original_img = preprocess_image(img_path)
    
    # 예측 수행
    output = model(image_tensor)
    target_class = torch.argmax(output).item()
    print(f'Predicted Class: {target_class}')
    
    # Grad-CAM 생성
    cam = grad_cam.generate_cam(image_tensor, target_class)
    overlayed_img = overlay_cam_on_image(original_img, cam)
    
    # 결과 시각화
    plt.figure(figsize=(10, 5))
    plt.subplot(1, 2, 1)
    plt.imshow(original_img)
    plt.title("Original Image")
    plt.axis('off')
    
    plt.subplot(1, 2, 2)
    plt.imshow(overlayed_img)
    plt.title("Grad-CAM")
    plt.axis('off')
    
    plt.show()

# 실행 예제
if __name__ == "__main__":
    img_path = "/HDD/datasets/visionsuite/cam/0016E5_07959.png"  # 실행할 이미지 파일
    main(img_path)